import re
import json
from pathlib import Path
from datetime import datetime
from typing import Dict, List, Optional
import os

class SECHumanReadableParser:
    def __init__(self, file_path):
        self.file_path = file_path
        # Tworzymy nową ścieżkę dla przetworzonego pliku
        base_dir = os.path.dirname(file_path)
        file_name = os.path.basename(file_path)
        self.output_path = os.path.join(
            base_dir, 
            'human_readable', 
            file_name.replace('.txt', '_human_readable.txt')
        )

    def _read_file(self) -> str:
        with open(self.file_path, 'r', encoding='utf-8') as f:
            return f.read()
            
    def _extract_header_info(self) -> Dict:
        """Wyciąga informacje z nagłówka SEC"""
        header_section = re.search(r'<SEC-HEADER>(.*?)</SEC-HEADER>', 
                                 self._read_file(), re.DOTALL)
        if not header_section:
            return {}
            
        header_text = header_section.group(1)
        return {
            'nazwa_firmy': self._find_value(header_text, r'COMPANY CONFORMED NAME:\s*(.+)'),
            'data_raportu': self._find_value(header_text, r'FILED AS OF DATE:\s*(.+)'),
            'typ_raportu': self._find_value(header_text, r'CONFORMED SUBMISSION TYPE:\s*(.+)'),
            'okres_raportu': self._find_value(header_text, r'CONFORMED PERIOD OF REPORT:\s*(.+)'),
            'adres': {
                'ulica': self._find_value(header_text, r'STREET 1:\s*(.+)'),
                'miasto': self._find_value(header_text, r'CITY:\s*(.+)'),
                'stan': self._find_value(header_text, r'STATE:\s*(.+)'),
                'kod': self._find_value(header_text, r'ZIP:\s*(.+)')
            }
        }

    def _find_value(self, text: str, pattern: str) -> str:
        match = re.search(pattern, text)
        return match.group(1).strip() if match else ''

    def _clean_html(self, text: str) -> str:
        """Usuwa tagi HTML i formatowanie XBRL"""
        text = re.sub(r'<[^>]+>', ' ', text)
        text = re.sub(r'\{.*?\}', '', text)
        text = re.sub(r'\s+', ' ', text)
        return text.strip()

    def generate_human_readable(self):
        """Generuje czytelną wersję raportu"""
        header_info = self._extract_header_info()
        
        output = []
        output.append("=" * 80)
        output.append(f"RAPORT SEC - WERSJA CZYTELNA")
        output.append("=" * 80 + "\n")
        
        # Informacje nagłówkowe
        output.append("PODSTAWOWE INFORMACJE:")
        output.append(f"Nazwa firmy: {header_info.get('nazwa_firmy')}")
        output.append(f"Data raportu: {header_info.get('data_raportu')}")
        output.append(f"Typ raportu: {header_info.get('typ_raportu')}")
        output.append(f"Okres raportu: {header_info.get('okres_raportu')}\n")
        
        # Adres
        adres = header_info.get('adres', {})
        output.append("ADRES SIEDZIBY:")
        output.append(f"{adres.get('ulica')}")
        output.append(f"{adres.get('miasto')}, {adres.get('stan')} {adres.get('kod')}\n")
        
        # Treść dokumentu
        output.append("TREŚĆ RAPORTU:")
        output.append("-" * 80)
        
        # Wyciągnij główną treść dokumentu
        main_content = re.search(r'<TEXT>(.*?)</TEXT>', self._read_file(), re.DOTALL)
        if main_content:
            cleaned_content = self._clean_html(main_content.group(1))
            output.append(cleaned_content)
            
        # Zapisz do pliku
        with open(self.output_path, 'w', encoding='utf-8') as f:
            f.write('\n'.join(output))
            
        return self.output_path

def process_report(file_path: str) -> str:
    """Funkcja pomocnicza do przetwarzania pojedynczego raportu"""
    parser = SECHumanReadableParser(file_path)
    return parser.generate_human_readable()

if __name__ == "__main__":
    # Przykład użycia
    report_path = "reports/0001385849/2024-10-16_8-K_0001062993-24-017723.txt"
    human_readable_path = process_report(report_path)
    print(f"Utworzono czytelną wersję raportu: {human_readable_path}") 